#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include <SDL/SDL_endian.h>
#include "IMG_savepng.h"

int wtiles=2, htiles=2, overlap=20;

#define OVERLAP_MAX 100

static struct
{
	const char *title;
	enum
	{
		TYPE_NONE,
		TYPE_STR,
		TYPE_INT
	} type;
	struct
	{
		int min, max;
	} range;

	void *dest;
} argtype[]={
	{ "-w", TYPE_INT,  {1, 65535}, &wtiles  },
	{ "-h", TYPE_INT,  {1, 65536}, &htiles  },
	{ "-l", TYPE_INT,  {0, OVERLAP_MAX/2  }, &overlap },
	{ NULL, TYPE_NONE, {0, 0    }, NULL     } };

enum
{
	ARG_WTILES,
	ARG_HTILES,
	ARG_OVERLAP
};

#define OVERLAP_P(X)	(((X)*overlap)/OVERLAP_MAX)
#define OVERLAP(X) (OVERLAP_P(X)+(X))

void usage(const char *executable)
{
	printf("Usage:\n\t%s [-w horizontal_tiles] [-h vertical_tiles]"
		" [-l overlap_percent] image.jpg [image2.jpg]\n", 
		executable);
}

int tile_pic(const char *fname)
{
	char title[512];
	SDL_Rect r;
	int x, y, tile_w, tile_h;
	SDL_Surface *s;

	strcpy(title, fname);
	{
		int end=strlen(title);
		while(end >= 0)
		{
			end--;
			if(title[end] == '.')
			{
				title[end]='\0';
				break;
			}
		}
	}

	s=IMG_Load(fname);

	if(s == NULL)
	{
		fprintf(stderr, "Couldn't load %s\n", fname);
		return(-1);
	}

	fprintf(stderr, "Loaded %s: %dx%d %dbpp\n", fname,
		s->w, s->h, s->format->BitsPerPixel);

	if(wtiles > s->w)
	{
		fprintf(stderr, "\tToo many tiles for width\n");
		goto TILE_CREATE_ERR;
	}

	if(htiles > s->h)
	{
		fprintf(stderr,"\tToo many tiles for height\n");
		goto TILE_CREATE_ERR;
	}

	tile_w=s->w/wtiles;
	tile_h=s->h/htiles;
	
	for(y=0; y<htiles; y++)
	{
		r.y=tile_h*y;
		r.h=tile_h;

		// TODO:  Calculate overlap on Y
		r.h+=OVERLAP_P(tile_h);
		if(y != 0)
			r.y -= (OVERLAP_P(tile_h)/2);

		for(x=0; x<wtiles; x++)
		{
			SDL_Surface *o;
			char name[512];
			sprintf(name, "%s_%02d_%02d.png", title, x, y);
			fprintf(stderr, "Creating tile %s...\n", name);

			r.x=tile_w*x;
			r.w=tile_w;
			r.w += OVERLAP_P(tile_w);

			if(x != 0)
				r.x -= OVERLAP_P(tile_w)/2;

			fprintf(stderr, "\tTile with overlap is"
				" {%d, %d: %dx%d}\n",
				r.x, r.y, r.w, r.h);

			o=SDL_CreateRGBSurface(SDL_SWSURFACE, r.w, r.h, 32,
				SDL_SwapLE32(0x000000ff),
				SDL_SwapLE32(0x0000ff00),
				SDL_SwapLE32(0x00ff0000),
				SDL_SwapLE32(0xff000000));

			SDL_BlitSurface(s, &r, o, NULL);
			IMG_SavePNG(name, o, IMG_COMPRESS_DEFAULT);
			SDL_FreeSurface(o);
		}
	}

	SDL_FreeSurface(s);
	return(0);

TILE_CREATE_ERR:
	SDL_FreeSurface(s);
	return(-1);
}

int main(int argc, char *argv[])
{
	int arg=1;

	if(SDL_Init(0) < 0)
	{
		fprintf(stderr, "Couldn't init SDL: %s\n", SDL_GetError());
		return(2);
	}
	else
		atexit(SDL_Quit);

	if(argc <= 1)
	{
		fprintf(stderr, "Must provide a filename\n");
		usage(argv[0]);
		return(1);
	}

	while(arg < argc)
	{
		int t;
		int arg_free=argc-arg;
		union
		{
			char str[512];
			int v;
		} value;

		if(argv[arg][0] != '-') break;

		for(t=0; argtype[t].title != NULL; t++)
		if(strcmp(argtype[t].title, argv[arg]) == 0)
		{
			if(argtype[t].type != TYPE_NONE)
			{
				if(arg_free < 1)
				{
					fprintf(stderr,
						"Need argument for %s\n", 
						argtype[t].title);
					arg++;
					break;
				}

				switch(argtype[t].type)
				{
				case TYPE_STR:
					if((long)strlen(argv[arg+1]) < argtype[t].range.min)
						fprintf(stderr, "String '%s' too short for %s\n", argv[arg+1],
							argtype[t].title);
					else if((long)strlen(argv[arg+1]) > argtype[t].range.max)
						fprintf(stderr, "String '%s' too long for %s\n", argv[arg+1],
							argtype[t].title);
					else
						strcpy(argtype[t].dest, argv[arg+1]);
					break;

				case TYPE_INT:
					if(sscanf(argv[arg+1], "%d", &value.v) != 1)
					{
						fprintf(stderr, "Invalid int argument %s for %s\n", argv[arg+1],
							argtype[t].title);
						break;
					}
					else if(value.v < argtype[t].range.min)
						fprintf(stderr, "Integer %d too small for %s\n", value.v, 
							argtype[t].title);
					else if(value.v > argtype[t].range.max)
						fprintf(stderr, "Integer %d too large for %s\n", value.v,
							argtype[t].title);
					else if(argtype[t].dest != NULL)
						(*(int *)(argtype[t].dest)) = value.v;

					break;

				default:
					fprintf(stderr, "Unknown arg type %d\n", argtype[t].type);
					break;
				}

				arg++;
			}

			arg++;

			break;
		}
	}

	fprintf(stderr, "wtiles=%d htiles=%d overlap=%d\n",
		wtiles, htiles, overlap);


	while(arg < argc)
		tile_pic(argv[arg++]);

	return(0);
}
